import dataclasses
import pathlib

import yuio.git
import yuio.io

if __name__ == "__main__":
    repo = yuio.git.Repo(pathlib.Path(__file__).parent.parent)

    status = repo.status()

    yuio.io.heading("Repository status")

    for k, v in dataclasses.asdict(status).items():
        if k != "changes":
            yuio.io.info("%s: `%s`", k, v)
        else:
            pass

    yuio.io.heading("Changes")

    if changes := status.changes:
        for change in changes:
            path = (
                change.path
                if change.path_from is None
                else f"{change.path_from} -> {change.path}"
            )
            yuio.io.info("%s: `%s%s`", path, change.staged.value, change.tree.value)
    else:
        yuio.io.info("No files were changed!")

    yuio.io.heading("Recent log")

    if log := repo.log(max_entries=5):
        for commit in log:
            yuio.io.info(
                "%s `[%s <%s>]`", commit.title, commit.author, commit.author_email
            )
    else:
        yuio.io.info("Log is empty!")



from __future__ import annotations

import enum
import functools
import json
import operator
import re
import typing as _t

import docutils.nodes
import sphinx.addnodes
import sphinx.application
from docutils.parsers.rst import directives
from sphinx.addnodes import desc_signature
from sphinx.domains.python import PyAttribute, PyClasslike, PyXRefRole
from sphinx.ext.autodoc import AttributeDocumenter, ClassDocumenter, bool_option
from sphinx.ext.autodoc.mock import ismock
from sphinx.util.docstrings import separate_metadata
from sphinx.util.inspect import isenumattribute, object_description

import yuio
import yuio.config
import yuio.json_schema

_TYPE_PARSE_RE = re.compile(
    r"""
    # Skip spaces, they're not meaningful in this context.
    \s+
    |
    (?P<dots>[.]{3})
    |
    # Literal string with escapes.
    # Example: `"foo"`, `"foo-\"-bar"`.
    (?P<string>(?P<string_q>['"`])(?:\\.|[^\\])*?(?P=string_q))
    |
    # Number with optional exponent.
    # Example: `1.0`, `.1`, `1.`, `1e+5`.
    (?P<number>(?:\d+(?:\.\d*)?|\.\d+)(?:[eE][+-]?\d+)?)
    |
    # Built-in type.
    # Example: `string`, `string?`.
    (?P<type>null|true|false|boolean|integer|number|string|any|never|object|set|array)\b
    \s*(?P<type_qm>\??)\s*
    |
    # Name component.
    (?P<name>\w[\w.#-]*)
    |
    # Punctuation that we separate with spaces.
    (?P<punct>[=:,|&])
    |
    # Punctuation that we copy as-is, without adding spaces.
    (?P<other_punct>[-!#$%()*+/;<>?@[\]^_{}~]+)
    |
    # Anything else is copied as-is.
    (?P<other>.)
    """,
    re.VERBOSE,
)


def type_to_nodes(typ: str, inliner) -> list[docutils.nodes.Node]:
    res = []

    for match in _TYPE_PARSE_RE.finditer(typ):
        if text := match.group("dots"):
            res.append(sphinx.addnodes.desc_sig_name(text, text))
        elif text := match.group("string"):
            res.append(sphinx.addnodes.desc_sig_literal_string(text, text))
        elif text := match.group("number"):
            res.append(sphinx.addnodes.desc_sig_literal_number(text, text))
        elif text := match.group("type"):
            res.append(sphinx.addnodes.desc_sig_keyword_type(text, text))
            if qm := match.group("type_qm"):
                res.append(sphinx.addnodes.desc_sig_punctuation(qm, qm))
        elif text := match.group("name"):
            text = "~" + text
            ref_nodes, warn_nodes = PyXRefRole(
                innernodeclass=sphinx.addnodes.desc_sig_name
            )("py:obj", text, text, 0, inliner)
            res.extend(ref_nodes)
            res.extend(warn_nodes)
        elif text := match.group("punct"):
            if text in "=|&":
                res.append(sphinx.addnodes.desc_sig_space())
            res.append(sphinx.addnodes.desc_sig_punctuation(text, text))
            res.append(sphinx.addnodes.desc_sig_space())
        elif text := match.group("other_punct"):
            res.append(sphinx.addnodes.desc_sig_punctuation(text, text))
        elif text := match.group("other"):
            res.append(docutils.nodes.Text(text))

    return res


def _get_fields(
    config: yuio.config.Config | type[yuio.config.Config],
) -> dict[str, yuio.config._Field]:
    return getattr(config, "_Config__get_fields")()


class Config(PyClasslike):
    option_spec = PyClasslike.option_spec.copy()
    option_spec["is-subconfig"] = directives.flag
    option_spec["display-name"] = directives.unchanged

    def handle_signature(self, sig: str, signode: desc_signature) -> tuple[str, str]:
        result = super().handle_signature(sig, signode)

        if canonical_name := self.options.get("display-name"):
            for node in signode.findall(sphinx.addnodes.desc_name):
                node.clear()
                node += docutils.nodes.Text(canonical_name)
            for node in signode.findall(sphinx.addnodes.desc_addname):
                node.parent.replace(node, [])

        return result

    def _object_hierarchy_parts(self, sig_node: desc_signature) -> tuple[str, ...]:
        if canonical_name := self.options.get("display-name"):
            return (canonical_name,)
        else:
            return super()._object_hierarchy_parts(sig_node)

    def _toc_entry_name(self, sig_node: desc_signature) -> str:
        if canonical_name := self.options.get("display-name"):
            return canonical_name
        else:
            return super()._toc_entry_name(sig_node)

    def get_signature_prefix(self, sig: str):
        if "is-subconfig" in self.options:
            return []
        else:
            return super().get_signature_prefix(sig)


class ConfigDocumenter(ClassDocumenter):
    objtype = "config"

    option_spec = ClassDocumenter.option_spec.copy()
    option_spec["enum-by-name"] = bool_option
    option_spec["enum-to-dash-case"] = bool_option
    option_spec["display-name"] = directives.unchanged

    @classmethod
    def can_document_member(
        cls, member: _t.Any, membername: str, isattr: bool, parent: _t.Any
    ):
        return (
            isinstance(parent, ConfigDocumenter)
            and isinstance(member, type)
            and issubclass(member, yuio.config.Config)
        )

    def format_signature(self, **kwargs) -> str:
        return ""

    def import_object(self, raiseerror: bool = False) -> bool:
        ret = super().import_object(raiseerror)
        if ret and (
            isinstance(self.parent, type)
            and issubclass(self.parent, yuio.config.Config)
        ):
            self.doc_as_attr = False
            self.is_nested_config = True
        else:
            self.is_nested_config = False
        return ret

    def add_directive_header(self, sig: str):
        super().add_directive_header(sig)
        if self.is_nested_config:
            self.add_line("   :is-subconfig:", self.get_sourcename())
        if self.options.display_name:
            self.add_line(
                "   :display-name: " + self.options.display_name, self.get_sourcename()
            )


class Field(PyAttribute):
    option_spec = PyAttribute.option_spec.copy()
    option_spec["display-name"] = directives.unchanged

    def handle_signature(self, sig: str, signode: desc_signature) -> tuple[str, str]:
        fullname, prefix = super(PyAttribute, self).handle_signature(sig, signode)

        if canonical_name := self.options.get("display-name"):
            for node in signode.findall(sphinx.addnodes.desc_name):
                node.clear()
                node += docutils.nodes.Text(canonical_name)

        typ = self.options.get("type")
        if typ:
            annotations = type_to_nodes(typ, self.state.inliner)
            signode += sphinx.addnodes.desc_annotation(
                typ,
                "",
                sphinx.addnodes.desc_sig_punctuation("", ":"),
                sphinx.addnodes.desc_sig_space(),
                *annotations,
            )

        value = self.options.get("value")
        if value:
            signode += sphinx.addnodes.desc_annotation(
                value,
                "",
                sphinx.addnodes.desc_sig_space(),
                sphinx.addnodes.desc_sig_punctuation("", "="),
                sphinx.addnodes.desc_sig_space(),
                docutils.nodes.Text(value),
            )

        return fullname, prefix

    def _object_hierarchy_parts(self, sig_node: desc_signature) -> tuple[str, ...]:
        parts = super()._object_hierarchy_parts(sig_node)
        if parts and (canonical_name := self.options.get("display-name")):
            parts = parts[:-1] + (canonical_name,)
        return parts

    def _toc_entry_name(self, sig_node: desc_signature) -> str:
        if not sig_node.get("_toc_parts"):
            return ""

        config = self.config
        *parents, name = sig_node["_toc_parts"]
        if config.toc_object_entries_show_parents == "domain":
            return name
        if config.toc_object_entries_show_parents == "hide":
            return name
        if config.toc_object_entries_show_parents == "all":
            return ".".join([*parents, name])
        return ""


class FieldDocumenter(AttributeDocumenter):
    objtype = "field"

    option_spec = AttributeDocumenter.option_spec.copy()
    option_spec["enum-by-name"] = bool_option
    option_spec["enum-to-dash-case"] = bool_option

    @classmethod
    def can_document_member(
        cls, member: _t.Any, membername: str, isattr: bool, parent: _t.Any
    ):
        return super().can_document_member(member, membername, isattr, parent) and (
            isinstance(parent, ConfigDocumenter)
        )

    def should_suppress_directive_header(self) -> bool:
        return True

    def add_directive_header(self, sig: str):
        super().add_directive_header(sig)
        sourcename = self.get_sourcename()
        if self.options.annotation:
            self.add_line("   :annotation: " + self.options.annotation, sourcename)
            return

        if not isinstance(self.parent, type):
            return
        if issubclass(self.parent, yuio.config.Config):
            if field := _get_fields(self.parent).get(self.object_name):
                self._add_header_for_field(field)
        elif issubclass(self.parent, enum.Enum):
            enumerator = getattr(self.parent, self.object_name, None)
            if enumerator and isenumattribute(enumerator):
                self._add_header_for_enum_field(enumerator)

    def _add_header_for_field(self, field: yuio.config._Field):
        if field.is_subconfig:
            return
        assert field.parser
        sourcename = self.get_sourcename()

        doc = self.get_doc() or []
        _, metadata = separate_metadata(
            "\n".join(functools.reduce(operator.iadd, doc, []))
        )
        if "hide-value" in metadata:
            return True

        if "type" in metadata:
            self.add_line("   :type: " + metadata["type"], sourcename)
        elif self.config.autodoc_typehints != "none":
            schema = field.parser.to_json_schema(
                yuio.json_schema.JsonSchemaContext()
            ).remove_opaque()
            if schema is not None:
                self.add_line("   :type: " + schema.pprint(), sourcename)

        if "value" in metadata:
            self.add_line("   :value: " + metadata["value"], sourcename)
        else:
            try:
                if (
                    self.options.no_value
                    or field.default is yuio.MISSING
                    or self.should_suppress_value_header()
                    or ismock(self.object)
                ):
                    pass
                else:
                    try:
                        objrepr = json.dumps(field.parser.to_json_value(field.default))
                    except TypeError:
                        objrepr = object_description(field.default)
                    self.add_line("   :value: " + objrepr, sourcename)
            except ValueError:
                pass

    def _add_header_for_enum_field(self, enumerator: enum.Enum):
        enum_by_name = self.options.enum_by_name
        enum_to_dash_case = self.options.enum_to_dash_case
        if enum_by_name:
            name = enumerator.name
        else:
            name = str(enumerator.value)
        if enum_to_dash_case:
            name = yuio.to_dash_case(name)
        self.add_line("   :display-name: " + name, self.get_sourcename())


def setup(app: sphinx.application.Sphinx):
    app.add_directive_to_domain("py", "config", Config)
    app.add_directive_to_domain("py", "field", Field)
    app.add_autodocumenter(ConfigDocumenter)
    app.add_autodocumenter(FieldDocumenter)

    return {
        "version": yuio.__version__,
        "parallel_read_safe": True,
        "parallel_write_safe": True,
    }



# Yuio project, MIT license.
#
# https://github.com/taminomara/yuio/
#
# You're free to copy this file to your project and edit it for your needs,
# just keep this copyright line please :3

"""
Controlling visual aspects of Yuio with themes.

Theme base class
----------------

The overall look and feel of a Yuio application is declared
in a :class:`Theme` object:

.. autoclass:: Theme

    .. autoattribute:: progress_bar_width

    .. autoattribute:: progress_bar_start_symbol

    .. autoattribute:: progress_bar_end_symbol

    .. autoattribute:: progress_bar_done_symbol

    .. autoattribute:: progress_bar_pending_symbol

    .. autoattribute:: spinner_pattern

    .. autoattribute:: spinner_update_rate_ms

    .. autoattribute:: spinner_static_symbol

    .. autoattribute:: msg_decorations

    .. automethod:: set_msg_decoration

    .. automethod:: _set_msg_decoration_if_not_overridden

    .. autoattribute:: colors

    .. automethod:: set_color

    .. automethod:: _set_color_if_not_overridden

    .. automethod:: get_color

    .. automethod:: to_color


Default theme
-------------

Use the following loader to create an instance of the default theme:

.. autofunction:: load

.. autoclass:: DefaultTheme

"""
from __future__ import annotations

import dataclasses
import functools
import os
from dataclasses import dataclass

import yuio.term
from yuio import _typing as _t
from yuio.term import Color, Term

T = _t.TypeVar("T")


class _ImmutableDictProxy(_t.Mapping[str, T], _t.Generic[T]):
    def __init__(self, data: dict[str, T], /, *, attr: str):
        self.__data = data
        self.__attr = attr

    def items(self) -> _t.ItemsView[str, T]:
        return self.__data.items()

    def keys(self) -> _t.KeysView[str]:
        return self.__data.keys()

    def values(self) -> _t.ValuesView[T]:
        return self.__data.values()

    def __len__(self):
        return len(self.__data)

    def __getitem__(self, key):
        return self.__data[key]

    def __iter__(self):
        return iter(self.__data)

    def __contains__(self, key):
        return key in self.__data

    def __repr__(self):
        return repr(self.__data)

    def __setitem__(self, key, item):
        raise RuntimeError(f"Theme.{self.__attr} is immutable")

    def __delitem__(self, key):
        raise RuntimeError(f"Theme.{self.__attr} is immutable")


class Theme:
    """Base class for Yuio themes."""

    _Self = _t.TypeVar("_Self", bound="Theme")

    #: Decorative symbols for certain text elements, such as headings,
    #: list items, etc.
    #:
    #: This mapping becomes immutable once a theme class is created. The only possible
    #: way to modify it is by using :meth:`~Theme.set_msg_decoration`
    #: or :meth:`~Theme._set_msg_decoration_if_not_overridden`.
    msg_decorations: _t.Mapping[str, str] = {
        "heading/section": "",
        "heading/1": "⣿ ",
        "heading/2": "",
        "heading/3": "",
        "heading/4": "",
        "heading/5": "",
        "heading/6": "",
        "question": "> ",
        "task": "> ",
        "thematic_break": "╌╌╌╌╌",
        "list": "•   ",
        "quote": ">   ",
        "code": " " * 8,
        # TODO: support these in widgets
        # 'menu_selected_item': '▶︎',
        # 'menu_default_item': '★',
        # 'menu_select': '#',
        # 'menu_search': '/',
    }

    #: An actual mutable version of :attr:`~Theme.msg_decorations`
    #: is kept here, because `__init_subclass__` will replace
    #: :attr:`~Theme.msg_decorations` with an immutable proxy.
    __msg_decorations: dict[str, str]
    #: Keeps track of where a message decoration was inherited from. This var is used
    #: to avoid `__init__`-ing message decorations that were overridden in a subclass.
    __msg_decoration_sources: dict[str, type | None] = {}

    #: Width of a progress bar for :class:`yuio.io.Task`.
    progress_bar_width: int = 15
    #: A symbol rendered on a left side of a progressbar.
    #:
    #: Set to ``"["`` to enclose progressbar in square brackets, for example.
    progress_bar_start_symbol: str = ""
    #: A symbol rendered on a right side of a progressbar.
    #:
    #: Set to ``"]"`` to enclose progressbar in square brackets, for example.
    progress_bar_end_symbol: str = ""
    #: Symbol rendered in the filled portion of a progressbar.
    progress_bar_done_symbol: str = "■"
    #: Symbol rendered in the unfilled portion of a progressbar.
    progress_bar_pending_symbol: str = "□"
    #: Spinner pattern for running tasks that don't have a progressbar.
    #:
    #: Every tick, a symbol in front of a task's heading updates, showing elements
    #: of this sequence.
    spinner_pattern: _t.Sequence[str] = "⣤⣤⣤⠶⠛⠛⠛⠶"
    #: How often the :attr:`~Theme.spinner_pattern` changes.
    spinner_update_rate_ms: int = 200
    #: Symbol for finished and failed tasks.
    #:
    #: It meant to resemble a static spinner.
    spinner_static_symbol = "⣿"

    #: Mapping of color paths to actual colors.
    #:
    #: Themes use color paths to describe styles and colors for different
    #: parts of an application. Color paths are similar to file paths,
    #: they use snake case identifiers separated by slashes, and consist of
    #: two parts separated by a colon.
    #:
    #: The first part represents an object, i.e. what are we coloring.
    #:
    #: The second part represents a context, i.e. what is the state or location
    #: of an object that we're coloring.
    #:
    #: For example, a color for the filled part of the task's progress bar
    #: has path ``"task/progressbar/done"``, a color for a text of an error
    #: log record has path ``"log/message:error"``, and a color for a string escape
    #: sequence in a highlighted python code has path ``"hl/str/esc:python"``.
    #:
    #: A color at a certain path is propagated to all sub-paths. For example,
    #: if ``"task/progressbar"`` is bold, and ``"task/progressbar/done"`` is green,
    #: the final color will be bold green.
    #:
    #: Each color path can be associated with either an instance
    #: of :class:`~yuio.term.Color`, another path, or a list of colors and paths.
    #:
    #: If path is mapped to a :class:`~yuio.term.Color`, then the path is associated
    #: with that particular color.
    #:
    #: If path is mapped to another path, then the path is associated with
    #: the color value for that other path (please don't create recursions here).
    #:
    #: If path is mapped to a list of colors and paths, then those colors and paths
    #: are combined.
    #:
    #: For example::
    #:
    #:     colors = {
    #:         'heading_color': Color.BOLD,
    #:         'error_color': Color.RED,
    #:         'tb/heading': ['heading_color', 'error_color'],
    #:     }
    #:
    #: Here, color of traceback's heading ``"tb/heading"`` will be bold and red.
    #:
    #: The base theme class provides colors for basic tags, such as `bold`, `red`,
    #: `code`, `note`, etc. :class:`DefaultTheme` expands on it, providing main
    #: colors that control the overall look of the theme, and then colors for all
    #: interface elements.
    #:
    #: When deriving from a theme, you can override this mapping. When looking up
    #: colors via :meth:`~Theme.get_color`, base classes will be tried for color,
    #: in order of method resolution.
    #:
    #: This mapping becomes immutable once a theme class is created. The only possible
    #: way to modify it is by using :meth:`~Theme.set_color`
    #: or :meth:`~Theme._set_color_if_not_overridden`.
    colors: _t.Mapping[str, str | Color | list[str | Color]] = {
        "code": "magenta",
        "note": "green",
        "bold": Color.STYLE_BOLD,
        "b": "bold",
        "dim": Color.STYLE_DIM,
        "d": "dim",
        "normal": Color.FORE_NORMAL,
        "normal_dim": Color.FORE_NORMAL_DIM,
        "red": Color.FORE_RED,
        "green": Color.FORE_GREEN,
        "yellow": Color.FORE_YELLOW,
        "blue": Color.FORE_BLUE,
        "magenta": Color.FORE_MAGENTA,
        "cyan": Color.FORE_CYAN,
    }

    #: An actual mutable version of :attr:`~Theme.colors`
    #: is kept here, because `__init_subclass__` will replace
    #: :attr:`~Theme.colors` with an immutable proxy.
    __colors: dict[str, str | Color | list[str | Color]]
    #: Keeps track of where a color was inherited from. This var is used
    #: to avoid `__init__`-ing colors that were overridden in a subclass.
    __color_sources: dict[str, type | None] = {}

    #: When running an `__init__` function, this variable will be set to the class
    #: that implemented it, regardless of type of `self`.
    #:
    #: That is, inside `DefaultTheme.__init__`, `__expected_source` is set
    #: to `DefaultTheme`, in `MyTheme.__init__` it is `MyTheme`, etc.
    #:
    #: This is possible because `__init_subclass__` wraps any implementation
    #: of `__init__` into a wrapper that sets this variable.
    __expected_source: type | None = None

    def __init__(self):
        pass

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)

        colors = {}
        color_sources = {}
        for base in reversed(cls.__mro__):
            base_colors = getattr(base, "_Theme__colors", {})
            colors.update(base_colors)
            base_color_sources = getattr(base, "_Theme__color_sources", {})
            color_sources.update(base_color_sources)

        colors.update(cls.colors)
        color_sources.update(dict.fromkeys(cls.colors.keys(), cls))

        cls.__colors = colors
        cls.__color_sources = color_sources
        cls.colors = _ImmutableDictProxy(cls.__colors, attr="colors")

        msg_decorations = {}
        msg_decoration_sources = {}
        for base in reversed(cls.__mro__):
            base_msg_decorations = getattr(base, "_Theme__msg_decorations", {})
            msg_decorations.update(base_msg_decorations)
            base_msg_decoration_sources = getattr(
                base, "_Theme__msg_decoration_sources", {}
            )
            msg_decoration_sources.update(base_msg_decoration_sources)

        msg_decorations.update(cls.msg_decorations)
        msg_decoration_sources.update(dict.fromkeys(cls.msg_decorations.keys(), cls))

        cls.__msg_decorations = msg_decorations
        cls.__msg_decoration_sources = msg_decoration_sources
        cls.msg_decorations = _ImmutableDictProxy(
            cls.__msg_decorations, attr="msg_decorations"
        )

        if init := cls.__dict__.get("__init__", None):

            @functools.wraps(init)
            def _wrapped_init(_self, *args, **kwargs):
                prev_expected_source = _self._Theme__expected_source
                _self._Theme__expected_source = cls
                try:
                    return init(_self, *args, **kwargs)
                finally:
                    _self._Theme__expected_source = prev_expected_source

            cls.__init__ = _wrapped_init  # type: ignore

    def _set_msg_decoration_if_not_overridden(
        self,
        name: str,
        msg_decoration: str,
        /,
    ):
        """Set message decoration by name, but only if it wasn't overridden
        in a subclass.

        This method should be called from `__init__` implementations
        to dynamically set message decorations. It will only set the decoration
        if it was not overridden by any child class.

        """

        if self.__expected_source is None:
            raise RuntimeError(
                f"_set_msg_decoration_if_not_overridden should only be called from __init__"
            )
        source = self.__msg_decoration_sources.get(name, Theme)
        # The class that's `__init__` is currently running should be a parent
        # of the msg_decoration's source. This means that the msg_decoration was assigned by a parent.
        if source is not None and issubclass(self.__expected_source, source):
            self.set_msg_decoration(name, msg_decoration)

    def set_msg_decoration(
        self,
        name: str,
        msg_decoration: str,
        /,
    ):
        """Set message decoration by name."""

        if "_Theme__msg_decorations" not in self.__dict__:
            self.__msg_decorations = self.__class__.__msg_decorations.copy()
            self.__msg_decoration_sources = (
                self.__class__.__msg_decoration_sources.copy()
            )
            self.msg_decorations = _ImmutableDictProxy(
                self.__msg_decorations, attr="msg_decorations"
            )
        self.__msg_decorations[name] = msg_decoration
        self.__msg_decoration_sources[name] = self.__expected_source

    def _set_color_if_not_overridden(
        self,
        path: str,
        color: str | Color | list[str | Color],
        /,
    ):
        """Set color by path, but only if the color was not overridden in a subclass.

        This method should be called from `__init__` implementations
        to dynamically set colors. It will only set the color if it was not overridden
        by any child class.

        """

        if self.__expected_source is None:
            raise RuntimeError(
                f"_set_color_if_not_overridden should only be called from __init__"
            )
        source = self.__color_sources.get(path, Theme)
        # The class who's `__init__` is currently running should be a parent
        # of the color's source. This means that the color was assigned by a parent.
        if source is not None and issubclass(self.__expected_source, source):
            self.set_color(path, color)

    def set_color(
        self,
        path: str,
        color: str | Color | list[str | Color],
        /,
    ):
        """Set color by path."""

        if "_Theme__colors" not in self.__dict__:
            self.__colors = self.__class__.__colors.copy()
            self.__color_sources = self.__class__.__color_sources.copy()
            self.colors = _ImmutableDictProxy(self.__colors, attr="colors")
        self.__colors[path] = color
        self.__color_sources[path] = self.__expected_source
        self.get_color.cache_clear()
        self.__dict__.pop("_Theme__color_tree", None)

    @dataclass(**yuio._with_slots())
    class __ColorTree:
        """
        Prefix-like tree that contains all of the theme's colors.

        """

        #: Colors in this node.
        colors: str | Color | list[str | Color] = Color.NONE

        #: Location part of the tree.
        loc: dict[str, Theme.__ColorTree] = dataclasses.field(default_factory=dict)

        #: Context part of the tree.
        ctx: dict[str, Theme.__ColorTree] = dataclasses.field(default_factory=dict)

    @functools.cached_property
    def __color_tree(self) -> Theme.__ColorTree:
        root = self.__ColorTree()

        for path, colors in self.__colors.items():
            loc, ctx = self.__parse_path(path)

            node = root

            for part in loc:
                if part not in node.loc:
                    node.loc[part] = self.__ColorTree()
                node = node.loc[part]

            for part in ctx:
                if part not in node.ctx:
                    node.ctx[part] = self.__ColorTree()
                node = node.ctx[part]

            node.colors = colors

        return root

    @staticmethod
    def __parse_path(path: str, /) -> tuple[list[str], list[str]]:
        path_parts = path.split(":", maxsplit=1)
        if len(path_parts) == 1:
            loc, ctx = path_parts[0], ""
        else:
            loc, ctx = path_parts
        return loc.split("/") if loc else [], ctx.split("/") if ctx else []

    @_t.final
    @functools.cache
    def get_color(self, path: str, /) -> Color:
        """Lookup a color by path."""

        loc, ctx = self.__parse_path(path)
        return self.__get_color_in_loc(self.__color_tree, loc, ctx)

    def __get_color_in_loc(
        self, node: Theme.__ColorTree, loc: list[str], ctx: list[str]
    ):
        color = Color.NONE

        for part in loc:
            if part not in node.loc:
                break
            color |= self.__get_color_in_ctx(node, ctx)
            node = node.loc[part]

        return color | self.__get_color_in_ctx(node, ctx)

    def __get_color_in_ctx(self, node: Theme.__ColorTree, ctx: list[str]):
        color = Color.NONE

        for part in ctx:
            if part not in node.ctx:
                break
            color |= self.__get_color_in_node(node)
            node = node.ctx[part]

        return color | self.__get_color_in_node(node)

    def __get_color_in_node(self, node: Theme.__ColorTree) -> Color:
        color = Color.NONE

        if isinstance(node.colors, str):
            color |= self.get_color(node.colors)
        elif isinstance(node.colors, list):
            for c in node.colors:
                color |= self.get_color(c) if isinstance(c, str) else c
        else:
            color |= node.colors

        return color

    def to_color(self, color_or_path: Color | str | None) -> Color:
        """
        Convert color or color path to color.

        """

        if color_or_path is None:
            return Color.NONE
        elif isinstance(color_or_path, Color):
            return color_or_path
        else:
            return self.get_color(color_or_path)


Theme.__init_subclass__()


class DefaultTheme(Theme):
    """Default Yuio theme. Adapts for terminal background color,
    if one can be detected.

    This theme defines *main colors*, which you can override by subclassing.

    - ``"heading_color"``: for headings,
    - ``"primary_color"``: for main text,
    - ``"accent_color"``: for visually highlighted elements,
    - ``"secondary_color"``: for visually dimmed elements,
    - ``"error_color"``: for everything that indicates an error,
    - ``"warning_color"``: for everything that indicates a warning,
    - ``"success_color"``: for everything that indicates a success,
    - ``"low_priority_color_a"``: for auxiliary elements such as help widget,
    - ``"low_priority_color_b"``: for auxiliary elements such as help widget,
      even lower priority.

    """

    #: Colors for default theme are separated into several sections.
    #:
    #: The main section (the first one) has common settings which are referenced
    #: from all other sections. You'll probably want to override
    colors = {
        #
        # Main settings
        # -------------
        # This section controls the overall theme look.
        # Most likely you'll want to change accent colors from here.
        "heading_color": ["bold", "primary_color"],
        "primary_color": "normal",
        "accent_color": "magenta",
        "accent_color_2": "cyan",
        "secondary_color": "normal_dim",
        "error_color": "red",
        "warning_color": "yellow",
        "success_color": "green",
        "low_priority_color_a": "normal_dim",
        "low_priority_color_b": "normal_dim",
        #
        # Common tags
        # -----------
        "code": "accent_color",
        "note": "accent_color_2",
        "path": "accent_color",
        #
        # IO messages and text
        # --------------------
        "msg/decoration": "secondary_color",
        "msg/decoration:heading": "accent_color",
        "msg/decoration:thematic_break": "secondary_color",
        "msg/text": "primary_color",
        "msg/text:heading": "heading_color",
        "msg/text:heading/section": "accent_color",
        "msg/text:question": "heading_color",
        "msg/text:error": "error_color",
        "msg/text:warning": "warning_color",
        "msg/text:success": ["bold", "success_color"],
        "msg/text:failure": ["bold", "error_color"],
        "msg/text:info": "primary_color",
        "msg/text:thematic_break": "secondary_color",
        #
        # Log messages
        # ------------
        "log/asctime": "secondary_color",
        "log/logger": "secondary_color",
        "log/level": "heading_color",
        "log/level:critical": "error_color",
        "log/level:error": "error_color",
        "log/level:warning": "warning_color",
        "log/level:info": "success_color",
        "log/level:debug": "secondary_color",
        "log/message": "primary_color",
        #
        # Tasks and progress bars
        # -----------------------
        "task": "secondary_color",
        "task/decoration:running": "accent_color",
        "task/decoration:done": "success_color",
        "task/decoration:error": "error_color",
        "task/progressbar/done": "accent_color",
        "task/progressbar/pending": "secondary_color",
        "task/heading": "heading_color",
        "task/progress": "secondary_color",
        "task/comment": "primary_color",
        #
        # Syntax highlighting
        # -------------------
        "hl/kwd": "bold",
        "hl/str": Color.NONE,
        "hl/str/esc": "bold",
        "hl/lit": Color.NONE,
        "hl/punct": "blue",
        "hl/punct:sh-usage": "secondary_color",
        "hl/comment": "secondary_color",
        "hl/prog": "bold",
        "hl/flag": "accent_color_2",
        "hl/meta:diff": "accent_color",
        "hl/added:diff": "green",
        "hl/removed:diff": "red",
        "hl/metavar": "bold",
        "tb/heading": ["bold", "red"],
        "tb/message": "tb/heading",
        "tb/frame/usr/file/module": "code",
        "tb/frame/usr/file/line": "code",
        "tb/frame/usr/file/path": "code",
        "tb/frame/usr/code": Color.NONE,
        "tb/frame/usr/highlight": "low_priority_color_a",
        "tb/frame/lib": "dim",
        "tb/frame/lib/file/module": "tb/frame/usr/file/module",
        "tb/frame/lib/file/line": "tb/frame/usr/file/line",
        "tb/frame/lib/file/path": "tb/frame/usr/file/path",
        "tb/frame/lib/code": "tb/frame/usr/code",
        "tb/frame/lib/highlight": "tb/frame/usr/highlight",
        #
        # Menu and widgets
        # ----------------
        "menu/text": "primary_color",
        "menu/text/heading": ["menu/text", "heading_color"],
        "menu/text/help_info:help": "low_priority_color_a",
        "menu/text/help_msg:help": "low_priority_color_b",
        "menu/text/help_key:help": "low_priority_color_a",
        "menu/text/help_sep:help": "low_priority_color_b",
        "menu/text/help_key:help_menu": "accent_color_2",
        "menu/text/help_sep:help_menu": "secondary_color",
        "menu/text/esc": "accent_color",
        "menu/text/comment": "accent_color_2",
        "menu/text/comment/decoration": "secondary_color",
        "menu/text:choice/active": "accent_color",
        "menu/text:choice/active/selected": ["bold"],
        "menu/text:choice/normal/selected": ["accent_color_2", "bold"],
        "menu/text:choice/normal/dir": "blue",
        "menu/text:choice/normal/exec": "red",
        "menu/text:choice/normal/symlink": "magenta",
        "menu/text:choice/normal/socket": "green",
        "menu/text:choice/normal/pipe": "yellow",
        "menu/text:choice/normal/block_device": ["cyan", "bold"],
        "menu/text:choice/normal/char_device": ["yellow", "bold"],
        "menu/text/comment:choice/normal/original": "success_color",
        "menu/text/comment:choice/normal/corrected": "error_color",
        "menu/text/prefix:choice/normal": "primary_color",
        "menu/text/prefix:choice/normal/selected": ["accent_color_2", "bold"],
        "menu/text/prefix:choice/active": "accent_color",
        "menu/text/prefix:choice/active/selected": ["bold"],
        "menu/text/suffix:choice/normal": "primary_color",
        "menu/text/suffix:choice/normal/selected": ["accent_color_2", "bold"],
        "menu/text/suffix:choice/active": "accent_color",
        "menu/text/suffix:choice/active/selected": ["bold"],
        "menu/text:choice/status_line": "low_priority_color_b",
        "menu/text:choice/status_line/number": "low_priority_color_a",
        "menu/placeholder": "secondary_color",
        "menu/decoration": "accent_color",
        "menu/decoration/quick-select": "secondary_color",
        "menu/decoration:choice/normal": "menu/text",
    }

    def __init__(self, term: Term):
        super().__init__()

        if term.terminal_colors is None:
            return

        # Gradients look bad in other modes.
        if term.has_colors_true:
            self._set_color_if_not_overridden(
                "task/progressbar/done/start", Color(fore=term.terminal_colors.blue)
            )
            self._set_color_if_not_overridden(
                "task/progressbar/done/end", Color(fore=term.terminal_colors.magenta)
            )

        if term.terminal_colors.lightness == yuio.term.Lightness.UNKNOWN:
            return

        background = term.terminal_colors.background
        foreground = term.terminal_colors.foreground

        if term.terminal_colors.lightness is term.terminal_colors.lightness.DARK:
            self._set_color_if_not_overridden(
                "low_priority_color_a",
                Color(fore=foreground.match_luminosity(background.lighten(0.30))),
            )
            self._set_color_if_not_overridden(
                "low_priority_color_b",
                Color(fore=foreground.match_luminosity(background.lighten(0.25))),
            )
        else:
            self._set_color_if_not_overridden(
                "low_priority_color_a",
                Color(fore=foreground.match_luminosity(background.darken(0.30))),
            )
            self._set_color_if_not_overridden(
                "low_priority_color_b",
                Color(fore=foreground.match_luminosity(background.darken(0.25))),
            )


def load(term: Term, /) -> Theme:
    """Loads a default theme."""

    # NOTE: loading themes from json is beta, don't use it yet.

    theme = DefaultTheme(term)

    if not (path := os.environ.get("YUIO_THEME_PATH")):
        return theme

    import yuio.config

    class ThemeData(yuio.config.Config):
        progress_bar_width: int | None = None
        progress_bar_start_symbol: str | None = None
        progress_bar_end_symbol: str | None = None
        progress_bar_done_symbol: str | None = None
        progress_bar_pending_symbol: str | None = None
        spinner_pattern: str | None = None
        spinner_update_rate_ms: int | None = None
        spinner_static_symbol: str | None = None
        msg_decorations: dict[str, str] | None = None
        colors: dict[str, str] | None = None

    theme_data = ThemeData.load_from_json_file(path)

    if theme_data.progress_bar_width is not None:
        theme.progress_bar_width = theme_data.progress_bar_width
    if theme_data.progress_bar_start_symbol is not None:
        theme.progress_bar_start_symbol = theme_data.progress_bar_start_symbol
    if theme_data.progress_bar_end_symbol is not None:
        theme.progress_bar_end_symbol = theme_data.progress_bar_end_symbol
    if theme_data.progress_bar_done_symbol is not None:
        theme.progress_bar_done_symbol = theme_data.progress_bar_done_symbol
    if theme_data.progress_bar_pending_symbol is not None:
        theme.progress_bar_pending_symbol = theme_data.progress_bar_pending_symbol
    if theme_data.spinner_pattern is not None:
        theme.spinner_pattern = theme_data.spinner_pattern
    if theme_data.spinner_update_rate_ms is not None:
        theme.spinner_update_rate_ms = theme_data.spinner_update_rate_ms
    if theme_data.spinner_static_symbol is not None:
        theme.spinner_static_symbol = theme_data.spinner_static_symbol

    if theme_data.msg_decorations is not None:
        for k, v in theme_data.msg_decorations.items():
            theme.set_msg_decoration(k, v)

    if theme_data.colors is not None:
        theme.colors = theme_data.colors

        for k, v in theme_data.colors.items():
            v = [Color.fore_from_hex(c) if c.startswith("#") else c for c in v.split()]
            theme.set_color(k, v[0] if len(v) == 1 else v)

    return theme



# Yuio project, MIT license.
#
# https://github.com/taminomara/yuio/
#
# You're free to copy this file to your project and edit it for your needs,
# just keep this copyright line please :3

"""
This module provides basic functionality to interact with git.
It comes in handy when writing deployment scripts.


Interacting with a repository
-----------------------------

All repository interactions are done through the :class:`Repo` class
and its methods. If an interaction fails, a :class:`GitException` is raised.

.. autoclass:: Repo
   :members:

.. autoclass:: GitException


Commit and status objects
-------------------------

Some of :class:`Repo` commands return parsed descriptions of git objects:

.. autoclass:: Commit
   :members:

.. autoclass:: CommitTrailers
   :members:

.. autoclass:: Status
   :members:

.. autoclass:: FileStatus
   :members:

.. autoclass:: Modification
   :members:


Parsing git refs
----------------

When you need to query a git ref from a user, :class:`RefParser` will check
that the given ref is formatted correctly. Use :class:`Ref` in your type hints
to help Yuio detect that you want to parse it as a git reference:

.. autoclass:: Ref

.. autoclass:: Tag

.. autoclass:: Branch

.. autoclass:: Remote

.. autoclass:: RefParser

.. autoclass:: TagParser

.. autoclass:: BranchParser

.. autoclass:: RemoteParser

If you know path to your repository before hand, and want to make sure that the user
supplies a valid ref that points to an existing git object, use :class:`CommitParser`:

.. autoclass:: CommitParser


Autocompleting git refs
-----------------------

.. autoclass:: RefCompleter

.. autoclass:: RefCompleterMode
    :members:

"""

from __future__ import annotations

import dataclasses
import enum
import functools
import logging
import pathlib
import re
import subprocess
import textwrap
from dataclasses import dataclass
from datetime import datetime

import yuio.complete
import yuio.parse
from yuio import _typing as _t

_logger = logging.getLogger(__name__)


class GitException(subprocess.SubprocessError):
    """Raised when git returns a non-zero exit code."""


#: A special kind of string that contains a git object reference.
#:
#: Ref is not guaranteed to be valid; this type is used in type hints
#: to make use of the :class:`RefParser`.
Ref = _t.NewType("Ref", str)

#: A special kind of string that contains a tag name.
#:
#: Ref is not guaranteed to be valid; this type is used in type hints
#: to make use of the :class:`TagParser`.
Tag = _t.NewType("Tag", str)

#: A special kind of string that contains a branch name.
#:
#: Ref is not guaranteed to be valid; this type is used in type hints
#: to make use of the :class:`BranchParser`.
Branch = _t.NewType("Branch", str)

#: A special kind of string that contains a remote branch name.
#:
#: Ref is not guaranteed to be valid; this type is used in type hints
#: to make use of the :class:`RemoteParser`.
Remote = _t.NewType("Remote", str)


_LOG_FMT = "%H%n%aN%n%aE%n%aI%n%cN%n%cE%n%cI%n%(decorate:prefix=,suffix=,tag=,separator= )%n%w(0,0,1)%B%w(0,0)%n-"
_LOG_TRAILERS_FMT = "%H%n%w(0,1,1)%(trailers:only=true)%w(0,0)%n-"
_LOG_TRAILER_KEY_RE = re.compile(r"^(?P<key>\S+):\s")


class Repo:
    """
    A class that allows interactions with a git repository.

    :param path:
        path to the repo root dir.
    :param skip_checks:
        don't check if we're inside a repo.

    """

    def __init__(self, path: pathlib.Path | str, /, skip_checks: bool = False):
        self.__path = pathlib.Path(path)

        try:
            self.git("--version")
        except FileNotFoundError:
            raise GitException(f"git executable was not found")

        if skip_checks:
            return

        try:
            self.git("rev-parse", "--is-inside-work-tree")
        except GitException:
            raise GitException(f"{path} is not a git repository")

    @property
    def path(self) -> pathlib.Path:
        """
        Path to the repo, as was passed to the constructor.

        """

        return self.__path

    def git(self, *args: str) -> bytes:
        """
        Call git and return its stdout.

        """

        if _logger.isEnabledFor(logging.DEBUG):
            _logger.debug("git %s", " ".join(args))

        res = subprocess.run(
            ["git"] + list(args),
            capture_output=True,
            cwd=self.__path,
        )

        if res.stderr:
            _logger.debug("%s", res.stderr.decode())

        if res.returncode != 0:
            raise GitException(
                f"git exited with status code {res.returncode}:\n"
                f"{res.stderr.decode()}"
            )

        return res.stdout

    def root(self) -> pathlib.Path:
        """
        Get the root directory of the repo.

        """

        return pathlib.Path(self.git("rev-parse", "--show-toplevel").decode())

    def git_dir(self) -> pathlib.Path:
        """
        Get path to the ``.git`` directory of the repo.

        """

        return pathlib.Path(self.git("rev-parse", "--git-dir").decode())

    def status(self) -> Status:
        """
        Query the current repository status.

        """

        text = self.git("status", "--porcelain=v2", "--branch", "-z")
        lines = iter(text.split(b"\0"))

        status = Status(commit="")

        for line_b in lines:
            line = line_b.decode()
            if line.startswith("# branch.oid"):
                status.commit = line[13:]
            elif line.startswith("# branch.head"):
                if line[14:] != "(detached)":
                    status.branch = line[14:]
            elif line.startswith("# branch.upstream"):
                status.upstream = line[18:]
            elif line.startswith("# branch.ab"):
                match = re.match(r"^\+(\d+) -(\d+)$", line[12:])
                assert match is not None
                status.ahead = int(match.group(1))
                status.behind = int(match.group(2))
            elif line.startswith("1"):
                match = re.match(r"^(.)(.) .{4} (?:[^ ]+ ){5}(.*)$", line[2:])
                assert match is not None
                file_status = FileStatus(
                    path=pathlib.Path(match.group(3)),
                    staged=Modification(match.group(1)),
                    tree=Modification(match.group(2)),
                )
                status.changes.append(file_status)
                status.has_tracked_changes |= (
                    file_status.staged is not Modification.UNTRACKED
                )
                status.has_untracked_changes |= (
                    file_status.staged is Modification.UNTRACKED
                )
            elif line.startswith("2"):
                match = re.match(r"^(.)(.) .{4} (?:[^ ]+ ){6}(.*)$", line[2:])
                assert match is not None
                file_status = FileStatus(
                    path=pathlib.Path(match.group(3)),
                    path_from=pathlib.Path(next(lines).decode()),
                    staged=Modification(match.group(1)),
                    tree=Modification(match.group(2)),
                )
                status.changes.append(file_status)
                status.has_tracked_changes |= (
                    file_status.staged is not Modification.UNTRACKED
                )
                status.has_untracked_changes |= (
                    file_status.staged is Modification.UNTRACKED
                )

        return status

    def log(self, *refs: str, max_entries: int | None = None) -> list[Commit]:
        """
        Query the log for given git objects.

        """

        args = [
            f"--pretty=format:{_LOG_FMT}",
            "--decorate-refs=refs/tags",
            "--decorate=short",
        ]

        if max_entries is not None:
            args += ["-n", str(max_entries)]

        args += list(refs)

        text = self.git("log", *args)
        lines = iter(text.decode().split("\n"))

        commits = []

        while commit := self.__parse_single_log_entry(lines):
            commits.append(commit)

        return commits

    def trailers(
        self, *refs: str, max_entries: int | None = None
    ) -> list[CommitTrailers]:
        """
        Query trailer lines for given git objects.

        Trailers are lines at the end of a commit message formatted as ``key: value``
        pairs. See `git-interpret-trailers`__ for further description.

        __ https://git-scm.com/docs/git-interpret-trailers

        """

        args = [f"--pretty=format:{_LOG_TRAILERS_FMT}"]

        if max_entries is not None:
            args += ["-n", str(max_entries)]

        args += list(refs)

        text = self.git("log", *args)
        lines = iter(text.decode().split("\n"))

        trailers = []

        while commit := self.__parse_single_trailer_entry(lines):
            trailers.append(commit)

        return trailers

    def show(self, ref: str, /) -> Commit | None:
        """
        Query information for the given git object.

        Return `None` if object is not found.

        """

        try:
            self.git("rev-parse", "--verify", ref)
        except GitException:
            return None

        text = self.git(
            "log",
            f"--pretty=format:{_LOG_FMT}",
            "-n1",
            ref,
        )

        lines = iter(text.decode().split("\n"))

        commit = self.__parse_single_log_entry(lines)

        if commit is None:
            return None

        commit.orig_ref = ref

        return commit

    @staticmethod
    def __parse_single_log_entry(lines) -> Commit | None:
        try:
            commit = next(lines)
            author = next(lines)
            author_email = next(lines)
            author_datetime = datetime.fromisoformat(next(lines))
            committer = next(lines)
            committer_email = next(lines)
            committer_datetime = datetime.fromisoformat(next(lines))
            tags = next(lines).split()
            title = next(lines)
            body = ""

            while True:
                line = next(lines)
                if not line or line.startswith(" "):
                    body += line[1:] + "\n"
                else:
                    break

            body = body.strip("\n")
            if body:
                body += "\n"

            return Commit(
                commit,
                tags,
                author,
                author_email,
                author_datetime,
                committer,
                committer_email,
                committer_datetime,
                title,
                body,
            )
        except StopIteration:
            return None

    @staticmethod
    def __parse_single_trailer_entry(
        lines,
    ) -> CommitTrailers | None:
        try:
            commit = next(lines)
            trailers = []
            current_key = None
            current_value = ""

            while True:
                line = next(lines)
                if not line or line.startswith(" "):
                    line = line[1:] + "\n"
                    if match := _LOG_TRAILER_KEY_RE.match(line):
                        if current_key:
                            first, *rest = current_value.splitlines(keepends=True)
                            current_value = (
                                first.strip() + "\n" + textwrap.dedent("".join(rest))
                            ).rstrip() + "\n"
                            trailers.append((current_key, current_value))
                        current_key = match.group("key")
                        current_value = line[match.end() :]
                    else:
                        current_value += line
                else:
                    break
            if current_key:
                first, *rest = current_value.splitlines(keepends=True)
                current_value = (
                    first.strip() + "\n" + textwrap.dedent("".join(rest))
                ).rstrip() + "\n"
                trailers.append((current_key, current_value))

            return CommitTrailers(commit, trailers)
        except StopIteration:
            return None

    def tags(self) -> list[str]:
        """
        List all tags in this repository.

        """

        return (
            self.git("for-each-ref", "--format=%(refname:short)", "refs/tags")
            .decode()
            .splitlines()
        )

    def branches(self) -> list[str]:
        """
        List all branches in this repository.

        """

        return (
            self.git("for-each-ref", "--format=%(refname:short)", "refs/heads")
            .decode()
            .splitlines()
        )

    def remotes(self) -> list[str]:
        """
        List all remote branches in this repository.

        """

        return (
            self.git("for-each-ref", "--format=%(refname:short)", "refs/remotes")
            .decode()
            .splitlines()
        )


@dataclass
class Commit:
    """
    Commit description.

    """

    #: Commit hash.
    hash: str

    #: Tags attached to this commit.
    tags: list[str]

    #: Author name.
    author: str

    #: Author email.
    author_email: str

    #: Author time.
    author_datetime: datetime

    #: Committer name.
    committer: str

    #: Committer email.
    committer_email: str

    #: Committer time.
    committer_datetime: datetime

    #: Commit title, i.e. first line of the message.
    title: str

    #: Commit body, i.e. the rest of the message.
    body: str

    #: If commit was parsed from a user input, this field will contain
    #: original input. I.e. if a user enters ``HEAD`` and it gets resolved
    #: into a commit, `orig_ref` will contain string ``"HEAD"``.
    #:
    #: See also :class:`CommitParser`.
    orig_ref: str | None = None

    @property
    def short_hash(self):
        """
        First seven characters of the commit hash.

        """

        return self.hash[:7]

    def __str__(self):
        if self.orig_ref:
            return self.orig_ref
        else:
            return self.short_hash


@dataclass
class CommitTrailers:
    """
    Commit trailers.

    """

    #: Commit hash.
    hash: str

    #: Key-value pairs for commit trailers.
    trailers: list[tuple[str, str]]


class Modification(enum.Enum):
    """For changed file, what modification was applied to it."""

    #: File wasn't changed.
    UNMODIFIED = "."

    #: File was changed.
    MODIFIED = "M"

    #: File was created.
    ADDED = "A"

    #: File was deleted.
    DELETED = "D"

    #: File was renamed (and possibly changed).
    RENAMED = "R"

    #: File was copied (and possibly changed).
    COPIED = "C"

    #: File with conflicts is unmerged.
    UNMERGED = "U"

    #: File is in ``.gitignore``.
    IGNORED = "?"

    #: File was created but not yet added to git, i.e. not staged.
    UNTRACKED = "!"


@dataclass
class FileStatus:
    """Status of a changed file."""

    #: Path of the file.
    path: pathlib.Path

    #: If file was moved, contains path where it was moved from.
    path_from: pathlib.Path | None = None

    #: File modification in the index (staged).
    staged: Modification = Modification.UNMODIFIED

    #: File modification in the tree (unstaged).
    tree: Modification = Modification.UNMODIFIED


@dataclass
class Status:
    """Status of a working copy."""

    #: Current commit hash.
    commit: str

    #: Name of the current branch.
    branch: str | None = None

    #: Name of the upstream branch.
    upstream: str | None = None

    #: Number of commits the branch is ahead of upstream.
    ahead: int | None = None

    #: Number of commits the branch is behind of upstream.
    behind: int | None = None

    #: True if any tracked file was changed.
    has_tracked_changes: bool = False

    #: True if any file was added but not tracked.
    has_untracked_changes: bool = False

    #: List of changed files, both tracked and untracked.
    changes: list[FileStatus] = dataclasses.field(default_factory=list)

    @property
    def has_changes(self) -> bool:
        """True if there are any changes in the repository."""

        return self.has_tracked_changes or self.has_untracked_changes


class RefCompleterMode(enum.Enum):
    """
    Specifies operation modes for :class:`RefCompleter`.

    """

    #: Completes branches.
    Branch = "b"

    #: Completes remote branches.
    Remote = "r"

    #: Completes tags.
    Tag = "t"

    #: Completes ``HEAD`` and ``ORIG_HEAD``.
    Head = "h"


class RefCompleter(yuio.complete.Completer):
    """
    Completes git refs.

    :param repo:
        source of completions.
    :param modes:
        which objects to complete.

    """

    def __init__(self, repo: Repo, modes: set[RefCompleterMode] | None = None):
        super().__init__()

        self._repo = repo
        self._modes = modes or {
            RefCompleterMode.Branch,
            RefCompleterMode.Tag,
            RefCompleterMode.Head,
        }

    def _process(self, collector: yuio.complete.CompletionCollector, /):
        try:
            if RefCompleterMode.Head in self._modes:
                collector.add_group()
                git_dir = self._repo.git_dir()
                for head in ["HEAD", "ORIG_HEAD"]:
                    if (git_dir / head).exists():
                        collector.add(head)
            if RefCompleterMode.Branch in self._modes:
                collector.add_group()
                for branch in self._repo.branches():
                    collector.add(branch, comment="branch")
            if RefCompleterMode.Remote in self._modes:
                collector.add_group()
                for remote in self._repo.remotes():
                    collector.add(remote, comment="remote")
            if RefCompleterMode.Tag in self._modes:
                collector.add_group()
                for tag in self._repo.tags():
                    collector.add(tag, comment="tag")
        except GitException:
            pass

    def _get_completion_model(
        self, *, is_many: bool = False
    ) -> yuio.complete._CompleterSerializer.Model:
        return yuio.complete._CompleterSerializer.Git(
            {
                yuio.complete._CompleterSerializer.Git.Mode(mode.value)
                for mode in self._modes
            }
        )


def CommitParser(*, repo: Repo) -> yuio.parse.Parser[Commit]:
    """
    A parser for git refs (commits, tags, branches, and so on).

    This parser validates that the given ref exists in the given repository,
    parses it and returns a commit data associated with this ref.

    If you need a simple string without additional validation,
    use :class:`RefParser`.

    """

    def map(value: str, /) -> Commit:
        commit = repo.show(value)
        if commit is None:
            raise yuio.parse.ParsingError("invalid git ref")
        return commit

    def rev(value: Commit | object) -> str:
        if isinstance(value, Commit):
            return str(value)
        elif isinstance(value, str):
            return value
        else:
            raise TypeError(
                f"parser Commit can't handle value "
                f"of type {_t.type_repr(type(value))}"
            )

    return yuio.parse.Map(yuio.parse.Str(), map, rev)


T = _t.TypeVar("T")


class _RefParserImpl(yuio.parse.ValidatingParser[T], _t.Generic[T]):
    if _t.TYPE_CHECKING:

        def __new__(
            cls,
            /,
            *,
            repo_path: Repo | str | pathlib.Path | None = None,
            should_exist: bool = False,
        ) -> _RefParserImpl[T]: ...

    def __init__(
        self,
        /,
        *,
        repo_path: Repo | str | pathlib.Path | None = None,
        should_exist: bool = False,
    ):
        super().__init__(_t.cast(yuio.parse.Parser[T], yuio.parse.Str()))

        if repo_path is None:
            repo_path = pathlib.Path.cwd()
        elif isinstance(repo_path, Repo):
            repo_path = repo_path.path
        else:
            repo_path = pathlib.Path(repo_path)

        self._repo_path = repo_path
        self._should_exist = should_exist

    @functools.cached_property
    def _repo(self) -> Repo:
        try:
            return Repo(self._repo_path, skip_checks=not self._should_exist)
        except GitException as e:
            raise yuio.parse.ParsingError(str(e)) from None

    def _validate(self, value: T, /):
        if self._should_exist and self._repo.show(str(value)) is None:
            raise yuio.parse.ParsingError(
                f"{value} does not exist in {self._repo_path}"
            )

    def completer(self) -> yuio.complete.Completer:
        return RefCompleter(self._repo)


class RefParser(_RefParserImpl[Ref]):
    """
    A parser that provides autocompletion for git refs, but doesn't verify
    anything else.

    """


class TagParser(_RefParserImpl[Tag]):
    """
    A parser that checks if the given string is a valid tag name.

    """

    def _validate(self, value: Tag, /):
        try:
            self._repo.git("check-ref-format", f"refs/tags/{value}")
        except GitException:
            raise yuio.parse.ParsingError(f"{value} is not a valid tag name") from None

        super()._validate(value)

    def completer(self) -> yuio.complete.Completer:
        return RefCompleter(self._repo, {RefCompleterMode.Tag})


class BranchParser(_RefParserImpl[Branch]):
    """
    A parser that checks if the given string is a valid tag name.

    """

    def _validate(self, value: Branch, /):
        try:
            self._repo.git("check-ref-format", f"refs/heads/{value}")
        except GitException:
            raise yuio.parse.ParsingError(
                f"{value} is not a valid branch name"
            ) from None

        super()._validate(value)

    def completer(self) -> yuio.complete.Completer:
        return RefCompleter(self._repo, {RefCompleterMode.Branch})


class RemoteParser(_RefParserImpl[Remote]):
    """
    A parser that checks if the given string is a valid tag remote branch.

    """

    def _validate(self, value: Remote, /):
        try:
            self._repo.git("check-ref-format", f"refs/remotes/{value}")
        except GitException:
            raise yuio.parse.ParsingError(
                f"{value} is not a valid branch name"
            ) from None

        super()._validate(value)

    def completer(self) -> yuio.complete.Completer:
        return RefCompleter(self._repo, {RefCompleterMode.Branch})


yuio.parse.register_type_hint_conversion(
    lambda ty, origin, args: yuio.parse._str_ty_union_parser(
        ty, origin, args, Ref, RefParser
    )
)

yuio.parse.register_type_hint_conversion(
    lambda ty, origin, args: yuio.parse._str_ty_union_parser(
        ty, origin, args, Tag, TagParser
    )
)

yuio.parse.register_type_hint_conversion(
    lambda ty, origin, args: yuio.parse._str_ty_union_parser(
        ty, origin, args, Branch, BranchParser
    )
)

yuio.parse.register_type_hint_conversion(
    lambda ty, origin, args: yuio.parse._str_ty_union_parser(
        ty, origin, args, Remote, RemoteParser
    )
)
